<?php
/**
 * SPDX-FileCopyrightText: 2024 Sascha Nitsch (grumpydeveloper) https://contentnation.net/@grumpydevelop
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * @author Yannis Vogel (vogeldevelopment)
 **/

namespace Federator\Api\FedUsers;

/**
 * handle activitypub inbox requests
 */
class Inbox implements \Federator\Api\FedUsers\FedUsersInterface
{
    /**
     * main instance
     *
     * @var \Federator\Api $main
     */
    private $main;

    /**
     * constructor
     * @param \Federator\Api $main api main instance
     */
    public function __construct($main)
    {
        $this->main = $main;
    }

    /**
     * handle get call
     *
     * @param string|null $_user user to fetch inbox for @unused-param
     * @return string|false response
     */
    public function get($_user)
    {
        return false;
    }

    /**
     * handle post call
     *
     * @param string|null $_user user to add data to inbox
     * @return string|false response
     */
    public function post($_user)
    {
        $_rawInput = file_get_contents('php://input');

        $allHeaders = getallheaders();
        try {
            $this->main->checkSignature($allHeaders);
        } catch (\Federator\Exceptions\PermissionDenied $e) {
            throw new \Federator\Exceptions\Unauthorized("Inbox::post Signature check failed: " . $e->getMessage());
        }

        $activity = is_string($_rawInput) ? json_decode($_rawInput, true) : null;

        if (!is_array($activity)) {
            throw new \Federator\Exceptions\ServerError("Inbox::post Input wasn't of type array");
        }

        $inboxActivity = \Federator\Data\ActivityPub\Factory::newActivityFromJson($activity);

        if ($inboxActivity === false) {
            throw new \Federator\Exceptions\ServerError("Inbox::post couldn't create inboxActivity");
        }

        $rootDir = PROJECT_ROOT . '/';

        // Shared inbox
        if (!isset($_user)) {
            // Save the raw input and parsed JSON to a file for inspection
            file_put_contents(
                $rootDir . 'logs/inbox.log',
                date('Y-m-d H:i:s') . ": ==== POST Inbox Activity ====\n" . json_encode($inboxActivity, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT) . "\n\n",
                FILE_APPEND
            );
        }

        $sendTo = $inboxActivity->getCC();
        if ($inboxActivity->getType() === 'Undo') { // for undo the object holds the proper cc
            $object = $inboxActivity->getObject();
            if ($object !== null && is_object($object)) {
                $sendTo = $object->getCC();
            }
        }

        $users = [];
        $dbh = $this->main->getDatabase();
        $cache = $this->main->getCache();
        $connector = $this->main->getConnector();

        foreach ($sendTo as $receiver) {
            if ($receiver === '' || !is_string($receiver)) {
                continue;
            }

            if (str_ends_with($receiver, '/followers')) {
                $actor = $inboxActivity->getAActor();
                if ($actor === null || !is_string($actor)) {
                    error_log("Inbox::post no actor found");
                    continue;
                }

                // Extract username from the actor URL
                $username = basename((string) (parse_url($actor, PHP_URL_PATH) ?? ''));
                $domain = parse_url($actor, PHP_URL_HOST);
                if ($username === null || $domain === null) {
                    error_log("Inbox::post no username or domain found");
                    continue;
                }
                $followers = \Federator\DIO\Followers::getFollowersByFedUser($dbh, $connector, $cache, $username . '@' . $domain);

                if (is_array($followers)) {
                    $users = array_merge($users, array_column($followers, 'id'));
                }
            }
        }
        if ($_user !== false && !in_array($_user, $users, true)) {
            $users[] = $_user;
        }
        foreach ($users as $user) {
            if (!isset($user)) {
                continue;
            }
            $token = \Resque::enqueue('inbox', 'Federator\\Jobs\\InboxJob', [
                'user' => $user,
                'activity' => $inboxActivity->toObject(),
            ]);
            error_log("Inbox::post enqueued job for user: $user with token: $token");
        }
        return "success";
    }

    /**
     * handle post call for specific user
     *
     * @param \mysqli $dbh database handle
     * @param \Federator\Connector\Connector $connector connector to use
     * @param \Federator\Cache\Cache|null $cache optional caching service
     * @param string $_user user to add data to inbox
     * @param \Federator\Data\ActivityPub\Common\Activity $inboxActivity the activity that we received
     * @return boolean response
     */
    public static function postForUser($dbh, $connector, $cache, $_user, $inboxActivity)
    {
        if (!isset($_user)) {
            error_log("Inbox::postForUser no user given");
            return false;
        }

        $user = \Federator\DIO\User::getUserByName(
            $dbh,
            $_user,
            $connector,
            $cache
        );
        if ($user === null || $user->id === null) {
            throw new \Federator\Exceptions\ServerError("Inbox::postForUser couldn't find user: $_user");
        }

        $rootDir = PROJECT_ROOT . '/';
        // Save the raw input and parsed JSON to a file for inspection
        file_put_contents(
            $rootDir . 'logs/inbox_' . $_user . '.log',
            date('Y-m-d H:i:s') . ": ==== POST " . $_user . " Inbox Activity ====\n" . json_encode($inboxActivity, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT) . "\n\n",
            FILE_APPEND
        );

        $type = $inboxActivity->getType();

        switch ($type) {
            case 'Follow':
                $success = false;
                $actor = $inboxActivity->getAActor();
                if ($actor !== '') {
                    $followerUsername = basename((string) (parse_url($actor, PHP_URL_PATH) ?? ''));
                    $followerDomain = parse_url($actor, PHP_URL_HOST);
                    if (is_string($followerDomain)) {
                        $followerId = "{$followerUsername}@{$followerDomain}";
                        $success = \Federator\DIO\Followers::addFollow($dbh, $followerId, $user->id, $followerDomain);
                    }
                }
                if ($success === false) {
                    error_log("Inbox::postForUser: Failed to add follower for user $user->id");
                }
                break;

            case 'Delete':
                // Delete Note/Post
                $object = $inboxActivity->getObject();
                if (is_string($object)) {
                    \Federator\DIO\Posts::deletePost($dbh, $object);
                }
                break;

            case 'Undo':
                $object = $inboxActivity->getObject();
                if (is_object($object) && method_exists($object, 'getType')) {
                    switch ($object->getType()) {
                        case 'Follow':
                            $success = false;
                            if ($object instanceof \Federator\Data\ActivityPub\Common\Activity) {
                                $actor = $object->getAActor();
                                if ($actor !== '') {
                                    $followerUsername = basename((string) (parse_url($actor, PHP_URL_PATH) ?? ''));
                                    $followerDomain = parse_url($actor, PHP_URL_HOST);
                                    if (is_string($followerDomain)) {
                                        $followerId = "{$followerUsername}@{$followerDomain}";
                                        $success = \Federator\DIO\Followers::removeFollow($dbh, $followerId, $user->id);
                                    }
                                }
                            }
                            if ($success === false) {
                                error_log("Inbox::postForUser: Failed to remove follower for user $user->id");
                            }
                            break;
                        case 'Like':
                            // Undo Like (remove like)
                            if (method_exists($object, 'getObject')) {
                                $targetId = $object->getObject();
                                if (is_string($targetId)) {
                                    // \Federator\DIO\Votes::removeVote($dbh, $user->id, $targetId, 'like');
                                    \Federator\DIO\Posts::deletePost($dbh, $targetId);
                                } else {
                                    error_log("Inbox::postForUser: Error in Undo Like for user $user->id, targetId is not a string");
                                }
                            }
                            break;
                        case 'Dislike':
                            // Undo Dislike (remove dislike)
                            if (method_exists($object, 'getObject')) {
                                $targetId = $object->getObject();
                                if (is_string($targetId)) {
                                    // \Federator\DIO\Votes::removeVote($dbh, $user->id, $targetId, 'dislike');
                                    \Federator\DIO\Posts::deletePost($dbh, $targetId);
                                } else {
                                    error_log("Inbox::postForUser: Error in Undo Dislike for user $user->id, targetId is not a string");
                                }
                            }
                            break;
                        case 'Note':
                            // Undo Note (remove note)
                            if (method_exists($object, 'getID')) {
                                $noteId = $object->getID();
                                \Federator\DIO\Posts::deletePost($dbh, $noteId);
                            }
                            break;
                    }
                }
                break;

            case 'Like':
                // Add Like
                $targetId = $inboxActivity->getObject();
                if (is_string($targetId)) {
                    // \Federator\DIO\Votes::addVote($dbh, $user->id, $targetId, 'like');
                    \Federator\DIO\Posts::savePost($dbh, $user->id, $inboxActivity);
                } else {
                    error_log("Inbox::postForUser: Error in Add Like for user $user->id, targetId is not a string");
                    return false;
                }
                break;

            case 'Dislike':
                // Add Dislike
                $targetId = $inboxActivity->getObject();
                if (is_string($targetId)) {
                    // \Federator\DIO\Votes::addVote($dbh, $user->id, $targetId, 'dislike');
                    \Federator\DIO\Posts::savePost($dbh, $user->id, $inboxActivity);
                } else {
                    error_log("Inbox::postForUser: Error in Add Dislike for user $user->id, targetId is not a string");
                    return false;
                }
                break;

            case 'Note':
                // Post Note
                \Federator\DIO\Posts::savePost($dbh, $user->id, $inboxActivity);
                break;

            default:
                error_log("Inbox::postForUser: Unhandled activity type $type for user $user->id");
                break;
        }

        return true;
    }
}
